#ifndef Yum_h
#define Yum_h

/*
** DANGER DANGER!  If you #include this, it must be #include'd first.  This is
** so that "Python.h" will be included first and so that its <Python.h> will be
** included first.  Python obnoxiously demands that that header be included
** first, and things sometimes go awry if it isn't.  (For example, you may get
** duplicate definitions of _POSIX_C_SOURCE.)
*/

#include "Python.h"
#include "Package.h"
#include "Log.h"
#include <QtGui/QWidget>
#include <QtCore/QHash>
#include <QtCore/QList>
#include <QtCore/QString>
#include <QtCore/QSet>
#include <QtCore/QEventLoop>
#include "Progress.h"
#include "YumIODialogs.h"
#include "YumIOCapturers.h"
#include "YumReposUsed.h"
#include <signal.h>

class CollectThread;
class ApplyThread;
class CollectFileListsThread;
class CollectOtherDataThread;
class ClearEntriesThread;

class YumIO : public QObject, private Python {
   Q_OBJECT
public:
   explicit YumIO();
   ~YumIO();
public:
   void  setup(QWidget*       parent,
               HProgressBar*  totalProgressBar,
               HProgressBar*  itemProgressBar);
public:
   bool  initialize             ();
   bool  collect                (const YumReposUsed&  reposUsed);
   bool  collectFileLists       ();
   bool  collectOtherData       ();
   void  clear                  ();
   void  clearEntries           ();  
   void  clearEntriesBeforeQuit ();
   bool  apply                  ();
public:
   bool  cleanHeaders      ();
   bool  cleanPackages     ();
   bool  cleanXmlMetadata  ();
   bool  cleanPickledCache ();
   bool  cleanSqliteCache  ();
   bool  cleanAll          ();
public:
   inline int  count() const 
   { if (clearing) return 0; else return list.count(); }
   inline Package::EntryList::const_iterator  begin() const
   { if (clearing) return emptyList.begin(); else return list.begin(); }
   inline Package::EntryList::const_iterator  end() const 
   { if (clearing) return emptyList.end(); else return list.end(); }
   inline Package::Entry* const&  operator[](int i) const
   { if (clearing) return emptyList[0]; else return list[i]; }
public:
   QString  getState            (const Package::Entry&  entry);
   QString  getKind             (const Package::Entry&  entry);
   // For packages & suites
   QString  getName             (const Package::Entry&  entry)
      { return entry.name; }
   QString  getVerRel           (const Package::Entry&  entry)
      { return entry.verrel; }
   QString  getRepo             (const Package::Entry&  entry)
      { return entry.repo; }
   QString  getArch             (const Package::Entry&  entry)
      { return entry.arch; }
   QString  getSize             (const Package::Entry&  entry)
      { return entry.sizeText; }
   QString  getSummary          (const Package::Entry&  entry)
      { return entry.summary; }
   QString  getDescription      (const Package::Entry&  entry);
   QString  getGroup            (const Package::Entry&  entry);
   QString  getVendor           (const Package::Entry&  entry);
   QString  getBuildTime        (const Package::Entry&  entry)
      { return entry.buildTimeText; }
   QString  getBuildHost        (const Package::Entry&  entry);
   QString  getInstallTime      (const Package::Entry&  entry);
   QString  getBinaryRpm        (const Package::Entry&  entry);
   QString  getSourceRpm        (const Package::Entry&  entry);
   QString  getLicense          (const Package::Entry&  entry);
   QString  getPackager         (const Package::Entry&  entry);
   QString  getUrl              (const Package::Entry&  entry);
   QString  getRelocations      (const Package::Entry&  entry);
   QString  getProvides         (const Package::Entry&  entry);
   QString  getRequires         (const Package::Entry&  entry);
   QString  getConflicts        (const Package::Entry&  entry);
   QString  getObsoletes        (const Package::Entry&  entry);
   QString  getFiles            (const Package::Entry&  entry);
   QString  getChangeLog        (const Package::Entry&  entry);
   // For groups
   QString  getRequiredGroups   (const Package::Entry&  entry);
   QString  getDefaultMetaPkgs  (const Package::Entry&  entry);
   QString  getOptionalMetaPkgs (const Package::Entry&  entry);
   QString  getMandatoryPkgs    (const Package::Entry&  entry);
   QString  getDefaultPkgs      (const Package::Entry&  entry);
   QString  getOptionalPkgs     (const Package::Entry&  entry);
   QString  getConditionalPkgs  (const Package::Entry&  entry);
public:
   bool     hasPickledCache ();
   // NOTE: See also usesSqlite().  2.3.2 - 2.9.1 support pickles or sqlite.
   bool     hasSqliteCache ();
   QString  getYumVar       (QString  name);
public:
   class EntryKey;
protected:
   typedef QHash<EntryKey,Package::Entry*>  EntryHash;
   Package::EntryList  list;
   Package::EntryList  emptyList;
   EntryHash           hash;
   QString             basearch;
   typedef QSet<QString>  StringSet;
   typedef QHash<QString,Package::Entry*>  EntryNameHash;
   EntryNameHash       namesBest;
   StringSet           biarches;
   bool                clearing;
protected:
   QStringList      collectRepos      (PyObject*              dict);
   void             useRepos          (PyObject*              repos,
                                       const YumReposUsed&    reposUsed);
   bool             collectField      (EntryHash&             hash,
                                       PyObject*              info,
                                       const char*            field,
                                       Package::State         state);
   bool             collectField      (EntryHash&             hash,
                                       PyObject*              info,
                                       char*                  field,
                                       Package::State         state);
   bool             collectSet        (EntryHash&             hash,
                                       PyObject*              set,
                                       Package::State         state);
   bool             collectPkg        (Package::Entry*&       entry,
                                       Package::State         state,
                                       PyObject*              pkg);
   bool             collectGroups     (EntryHash&             hash,
                                       PyObject*              groupInfo);
   bool             collectGroup      (Package::Entry*&       entry,
                                       PyObject*              groupObj,
                                       QString                name);
   void             clearEntriesReal  ();
   bool             upToDateGroup     (const Package::Entry&  entry);
   bool             upToDateGroupGrps (const Package::Entry&  entry,
                                       const char*            attr_22);
   bool             upToDateGroupGrps (const Package::Entry&  entry,
                                       char*                  attr_22);
   bool             upToDateGroupPkgs (const Package::Entry&  entry,
                                       const char*            attr_22,
                                       const char*            attr_25);
   bool             upToDateGroupPkgs (const Package::Entry&  entry,
                                       char*                  attr_22,
                                       char*                  attr_25);
   QString          getPrco           (PyObject*              pkg,
                                       const char*            hdrKey,
                                       const char*            prcoKey);
   QString          getPrco           (PyObject*              pkg,
                                       const char*            hdrKey,
                                       char*                  prcoKey);
#if 0
   QString          getRpm            (PyObject*              pkg,
                                       char*                  rpmKey);
#endif
   QString          getRpmTime        (PyObject*              pkg,
                                       const char*            rpmKey);
   QString          getRpmList        (PyObject*              pkg,
                                       const char*            rpmKey);
   QString          getGroupGrps      (const Package::Entry&  entry,
                                       const char*            attr_22);
   QString          getGroupGrps      (const Package::Entry&  entry,
                                       char*                  attr_22);
   QString          getGroupPkgs      (const Package::Entry&  entry,
                                       const char*            attr_22,
                                       const char*            attr_25);
   QString          getGroupPkgs      (const Package::Entry&  entry,
                                       char*                  attr_22,
                                       char*                  attr_25);
   QString          makeSorted        (QStringList&           list);
   QString          makeSortedUniq    (QStringList&           list);
   PyObject*        getRpmHdrValue    (PyObject*              pkg,
                                       const char*            hdrKey);
   QString          getLocalPkg       (PyObject*              pkg);
   QString          getLocalHdr       (PyObject*              pkg);
   Package::Kind    getPkgKind        (PyObject*              pkg);
   PyObject*        getGroupItem_22   (const Package::Entry&  entry,
                                       char*                  attr);
   PyObject*        getGroupItem_25   (const Package::Entry&  entry,
                                       char*                  attr);
   Package::Entry*  findPackage       (QString                name);
   Package::Entry*  findPackageAction (QString                name,
                                       Package::Action        action);
   bool             parseEnvra        (QString                nevra,
                                       QString&               name,
                                       QString&               epoch,
                                       QString&               version,
                                       QString&               release,
                                       QString&               arch);
   bool             parseEnvra        (QString                nevra,
                                       QString&               name,
                                       QString&               arch);
   bool             applyGroups       ();
   bool             applyGrpInstall   (const Package::Entry*  grp);
   bool             applyGrpUpdate    (const Package::Entry*  grp);
   bool             applyGrpRemove    (const Package::Entry*  grp);
   bool             applyGrpConflict  (const Package::Entry*  grp,
                                       Package::Entry*        pkg);
   void             applyGrpLog       (const Package::Entry*  grp,
                                       const Package::Entry*  pkg);
   void             applyGrpLogWas    (const Package::Entry*  grp,
                                       const Package::Entry*  pkg);
   void             applyGrpLogNot    (const Package::Entry*  grp,
                                       const Package::Entry*  pkg,
                                       Package::Action        action,
                                       QString                reason);
   PyObject*        requiredPkgs      (const Package::Entry*  grp);
   PyObject*        pkgTree           (const Package::Entry*  grp);
   PyObject*        allPkgs           (const Package::Entry*  grp);
   bool             makeLists         (PyObject*              yumIO,
                                       PyObject*&             tsInfo,
                                       PyObject*&             lists,
                                       PyObject*&             updated,
                                       PyObject*&             installed,
                                       PyObject*&             removed,
                                       PyObject*&             obsoleted,
                                       PyObject*&             depup,
                                       PyObject*&             depinst,
                                       PyObject*&             deprem);
   QString          simpleString      (PyObject*              pkg,
                                       const char*            key);
   QString          simpleString      (PyObject*              pkg,
                                       char*                  key);
   QString          getBiArch         (QString                arch);
   Package::Size    sizeOf            (PyObject*              pkg);
   QString          bytesFormat       (Package::Size          bytes);
   time_t           buildTimeOf       (PyObject*              pkg);
   char*            dateFormat        (time_t                 secs);
   char*            timeFormat        (time_t                 secs);
   QString          yumVersionStr     ();
   long             yumVersion        ();
   bool             hasDistribGroups  ();
   bool             hasTransTest      ();
   bool             usesSqlite        ();
   bool             usesPresto        ();
   bool             hasMultiProgress  ();
   QString          pyString          (PyObject*              obj);
   QString          pyStringDec       (PyObject*              obj);
   QString          pyStringX         (PyObject*              obj);
   QString          pyStringXDec      (PyObject*              obj);
   QString          pyMappingString   (PyObject*              obj,
                                       const char*            attr);
   time_t           pyTime            (PyObject*              obj);
   time_t           pyTimeDec         (PyObject*              obj);
   PyObject*        pyDictGetItem     (PyObject*              dict,
                                       const QString&         key);
   const char*      pyDescribe        (PyObject*              obj);
   void             log               (Log::Kind              kind,
                                       unsigned int           value,
                                       QString                text);
   bool             reapChildren      ();
   void             sigchldDisable    ();
   void             sigchldRestore    ();
   virtual void     pyExceptExpect    (PyObject*              xcpt);
   virtual void     pyExceptClear     ();
   void             pythonException   (QString                text);
   void             pythonExceptionV  (QString                format);
   void             commExceptSetup   ();
   bool             commExceptRaised  (PyObject*              result,
                                       QString                descr);
protected slots:
   void  blockerRelease ();
protected:
   void  execWorkThread       (QEventLoop&          eventLoop,
                               QThread&             thread);
   bool  collectReal          (const YumReposUsed&  reposUsed);
   bool  applyReal            ();
   bool  collectFileListsReal ();
   bool  collectOtherDataReal ();
   friend class CollectThread;
   friend class ApplyThread;
   friend class CollectFileListsThread;
   friend class CollectOtherDataThread;
   friend class ClearEntriesThread;
public:
   // These are public because they're callbacks from YumIOC.
   PyObject*  stdoutWrite                 (PyObject*  self, PyObject*  args);
   PyObject*  stderrWrite                 (PyObject*  self, PyObject*  args);
   PyObject*  log                         (PyObject*  self, PyObject*  args);
   PyObject*  errorlog                    (PyObject*  self, PyObject*  args);
   PyObject*  collectProgress             (PyObject*  self, PyObject*  args);
   PyObject*  collectGrabMeterStart       (PyObject*  self, PyObject*  args);
   PyObject*  collectGrabMeterUpdate      (PyObject*  self, PyObject*  args);
   PyObject*  collectGrabMeterEnd         (PyObject*  self, PyObject*  args);
   PyObject*  collectGrabMeterFailure     (PyObject*  self, PyObject*  args);
   PyObject*  collectGrabMultiMeterStart  (PyObject*  self, PyObject*  args);
   PyObject*  collectGrabMultiMeterNew    (PyObject*  self, PyObject*  args);
   PyObject*  collectGrabMultiMeterRemove (PyObject*  self, PyObject*  args);
   PyObject*  collectGrabMultiMeterEnd    (PyObject*  self, PyObject*  args);
   PyObject*  applyGrabMeterStart         (PyObject*  self, PyObject*  args);
   PyObject*  applyGrabMeterUpdate        (PyObject*  self, PyObject*  args);
   PyObject*  applyGrabMeterEnd           (PyObject*  self, PyObject*  args);
   PyObject*  applyGrabMultiMeterStart    (PyObject*  self, PyObject*  args);
   PyObject*  applyGrabMultiMeterNew      (PyObject*  self, PyObject*  args);
   PyObject*  applyGrabMultiMeterRemove   (PyObject*  self, PyObject*  args);
   PyObject*  applyGrabMultiMeterEnd      (PyObject*  self, PyObject*  args);
   PyObject*  dsTsCheck                   (PyObject*  self, PyObject*  args);
   PyObject*  dsTsLog                     (PyObject*  self, PyObject*  args);
   PyObject*  dsRestartLoop               (PyObject*  self, PyObject*  args);
   PyObject*  dsEnd                       (PyObject*  self, PyObject*  args);
   PyObject*  userconfirm                 (PyObject*  self, PyObject*  args);
   PyObject*  reportDownloadSize          (PyObject*  self, PyObject*  args);
   PyObject*  listTransaction             (PyObject*  self, PyObject*  args);
   PyObject*  tsInitCallback              (PyObject*  self, PyObject*  args);
   PyObject*  tsCallback                  (PyObject*  self, PyObject*  args);
   PyObject*  tsEvent                     (PyObject*  self, PyObject*  args);
   PyObject*  postTransactionOutput       (PyObject*  self, PyObject*  args);
protected:
   enum RepoAct {
      RA_Grab_Metalink,
      RA_Grab_RepoMD,
      RA_Grab_Group,
      RA_Grab,
      RA_Convert,
      RA_Read,
      RA_Count,        // number of acts for each repository
   };
   void  collectStart   (void*    meterObj,
                         QString  name,
                         RepoAct  act,
                         long     current,
                         long     total);
   void  collectUpdate  (void*    meterObj,
                         long     current);
   void  collectEnd     (void*    meterObj,
                         long     current);
   void  collectFailure (void*    meterObj);
public:
   void             writeStdout                (QString          content);
   void             writeStderr                (QString          content);
protected:
   void             dsTsInfo                   (char*            msg);
   Package::Entry*  dsTsFindPkg                (QString          name);
   void             dsTsNewPkg                 (QString          name,
                                                Package::Entry*  entry,
                                                unsigned int     numTs,
                                                bool             needHdr);
   PyObject*        write                      (PyObject*        self,
                                                PyObject*        args,
                                                Log::Kind        kind,
                                                unsigned int     value,
                                                QString&         buffer);
   void             write                      (QString          content,
                                                Log::Kind        kind,
                                                unsigned int     value,
                                                QString&         buffer);
   void             exceptionExpected          ();
protected:
   QString          pkgtupFormat               (PyObject*        pkgtup);
   QString          pkgtupRepoFormat           (PyObject*        pkgtup,
                                                QString          repoid);
   int              displayTransactionPkgs     (QString          action,
                                                PyObject*        txmbrs);
   int              displayTransactionPkgsRepo (QString          action,
                                                PyObject*        txmbrs);
   PyObject*        findRelation               (PyObject*        relatedto,
                                                const char*      relation);
   int              displayTransactionPkgsObs  (QString          action,
                                                PyObject*        txmbrs);
   void             noteTransactionPkgs        (Package::Action  action,
                                                PyObject*        txmbrs);
signals:
   void             collectedExtraData();
protected:
   void error   (const char*  text);
   void warning (const char*  text);
public:
   void error   (QString  text);
   void warning (QString  text);
public:
   Log::EntryList     logEntries;
   int                logStartOp;
   bool               collectedFileLists;
   bool               collectedOtherData;
protected:
   QWidget*           parent;
   YumIODialogs*      dialogs;
   YumIOCapturers*    capturers;
   PyObject*          rpm_dict;
   PyObject*          classYum;
   PyObject*          yumIO;
   PyObject*          reposDict;
   bool               locked;
   bool               yumVerKnown;
   QString            yumVer;
   bool               usesSqliteKnown;
   bool               usesSqliteValue;
   bool               usesPrestoKnown;
   bool               usesPrestoValue;
   bool               hasMultiProgressKnown;
   bool               hasMultiProgressValue;
   struct {
      PyObject*       bdbQuit;
      PyObject*       yumBaseError;
      PyObject*       repoError;
      PyObject*       sqlite3OperationalError;
      PyObject*       lockError;
      PyObject*       groupsError;
      PyObject*       noMoreMirrorsRepoError;
   } xcpt;
   struct {
      QString         stdoutBuffer;
      QString         stderrBuffer;
      Log::Kind       stderrKind;
      unsigned int    stderrValue;
   } wr;
   HProgressBar*      totalProgressBar;
   HProgressBar*      itemProgressBar;
   Progress::Object*  pg;
   // Progress for reading repositories (collect)
   struct ThreadId {
      QString  name;
      RepoAct  act;
      explicit ThreadId() {}
      explicit ThreadId(QString name, RepoAct act) : name(name), act(act) {}
      operator long int() const { return qHash(name) + act; }
   };
   typedef  QHash<void*,Progress::Threads>     MeterMapping;
   typedef  QHash<ThreadId,Progress::Threads>  RetryMapping;
   struct {
      QStringList  repos;
      QString      kind;
      int          act;   // There are RepoAct::Count acts per repo
      int          total;
      QString      name;
      struct {
         bool               begun;
         bool               setup;
         long long          files;
         long long          size;
         Progress::Threads  threadsStarted;
         MeterMapping       meterMapping;
         ThreadId*          threads;
         RetryMapping       retryMapping;
      } parallel;
   } repoProgress;
   // State for transaction (apply)
   struct {
      QString        descr;
      bool           confirmed;
      unsigned int   act;  
      // State for checking dependencies (including hdr grabs)
      struct {
         unsigned int  totalTs;           // number of installs & erases
         unsigned int  totalGrabs;        // number of headers to download
         unsigned int  depSolveActs;      // depsolve acts for this iteration
         unsigned int  totalDepSolveActs; // depsolve acts for whole check
      } dep;
      // State for running transaction (including rpm grabs)
      struct {
         Package::Size  nominalGrabSize; // size of rpm's needed
         Package::Size  totalGrabSize;   // size of rpm's not in cache
         unsigned int   pgRGrabRpms;     // total progress for rpm grabs
         Package::Size  totalTsSize;     // size of installs & erases
         unsigned long  lastTs;          // last install/erase
         bool           deltaRebuilt;    // <delta rebuild> happened
      } run;
      // State for parallel grabs (either hdr or rpm grabs, depending on
      // ts.confirmed).
      struct {
         bool                            begun;
         long long                       files;
         Package::Size                   size;
         Progress::Threads               threadsStarted;
         QHash<void*,Progress::Threads>  meterMapping;
      } parallel;
   } ts;
   struct sigaction  savedSigChldAct;
   class Blocker : QMutex {
   public:
      Blocker() : QMutex() {}
      void prepare() { lock(); }           /* called by creator thread */
      void release() { unlock(); }         /* called by creator thread */
      void block()   { lock(); unlock(); } /* called by created thread */
   };
   Blocker blocker;
};

#endif
